#ifndef __AUTOMGM_H__
#define __AUTOMGM_H__
#pragma once

/// Simple useful classes for dealing with allocated character strings
struct achar {
  char *c;
  DWORD s, l;
  operator char*() { return c;}
  operator unsigned char*() { return (unsigned char*)c;}
  operator const char*() const { return c;}
  operator const unsigned char*() const { return (unsigned char*)c;}
  achar() { c = 0; s = 0; l = 0; }
  achar(DWORD n) { c = (char*)malloc(s = ++n); l = 0; }
  achar(const achar &a) { s = a.s, l = a.l; c = a.c? (char*)malloc(s) : 0; c && memcpy(c, a.c, l + 1); }
  void operator =(const achar &a) { s = a.s, l = a.l; c = a.c? (char*)realloc(c, s) : (free(c),0); c && memcpy(c, a.c, l + 1); }
  achar(const char *ss) { c = 0; if (ss) strcpy(resize(l = strlen(ss)), ss); else (s = l = 0); }
  ~achar() { free(c); }
  char *resize(DWORD n) { c = (char*)realloc(c, s = ++n); return c; }
  char *resize() { return l >= s? resize(l) : c; }
  void checklen() { l = c? strlen(c) : 0; }
  DWORD asize() const { return s; }
  DWORD size() const { return l; }
  void strlwr() { for(char *s = c, *e = c + l; s < e; s++) *s = tolower(*s); }
  UINT GetDlgItemText(HWND hwnd, int nIDDlgItem);
  UINT GetDlgItemTextUnCEsc(HWND hwnd, int nIDDlgItem);
  LONG QueryValue(HKEY hk, const char *name, DWORD &type);
};

struct fchar {
  char *c;
  fchar() : c(0) {}
  explicit fchar(void *v) : c((char*)v) {}
  fchar(fchar &f) : c(f.c) { f.c = 0; }
  operator char*&() { return c;}
  operator unsigned char*&() { return (unsigned char*&)c;}
  char &operator *() { return *c; }
  operator const char*() const { return c;}
  operator const unsigned char*() const { return (unsigned char*)c;}
  //    operator bool() const { return c != 0; }
  ~fchar() { free(c); }
private:
//  fchar(const fchar &f);
};


struct auto_close_handle {
  HANDLE h;
  auto_close_handle(HANDLE H) : h(H) {}
  ~auto_close_handle() { CloseHandle(h); }
};

struct auto_close_hkey {
  HKEY hk;
  auto_close_hkey(HKEY HK) : hk(HK) {}
  ~auto_close_hkey() { if (hk != (HKEY)INVALID_HANDLE_VALUE) RegCloseKey(hk); }
};

#endif //__AUTOMGM_H__
